package test;

import java.math.BigDecimal;
import java.math.RoundingMode;
import java.util.Iterator;

import valueObjects.ActualValue;

import containers.ActualValueContainer;
import containers.AlabamaEnrollmentsContainer;
import containers.TrapezoidalFuzzyNumberContainer;
import containers.FuzzySetContainer;
import containers.TrapezoidalFuzzySetContainer;
import fuzzySet.FuzzyNumber;
import fuzzySet.FuzzySet;
import fuzzySet.TrapezoidalFuzzyNumber;
import fuzzySet.TrapezoidalFuzzySet;

public class TrapezoidalFuzzifier {

	private FuzzySetContainer fuzzySets = new TrapezoidalFuzzySetContainer();
	private TrapezoidalFuzzyNumberContainer fuzzyNumbers = new TrapezoidalFuzzyNumberContainer();
	ActualValueContainer actualValues;
	private double valueDifferences[];

	private int id = 1;

	public TrapezoidalFuzzifier(ActualValueContainer actualValues) {
		this.actualValues = actualValues;
		init();
	}

	public Iterator<FuzzyNumber> getFuzzyNumberContainer() {
		return fuzzyNumbers.getFuzzyNumberContainerIterator();
	}

	private void init() {
		initValueDifferencesArray();
		initFuzzyNumbers();
		initFuzzySets();
	}

	public FuzzySetContainer getFuzzySets() {
		return fuzzySets;
	}

	public Iterator<FuzzySet> getFuzzySetsIterator() {
		return fuzzySets.getFuzzySetsIterator();
	}

	public FuzzySet getLastFuzzySet() {
		return fuzzySets.getFuzzySet(fuzzySets.getNumberOfSets() - 1);
	}

	public int getNumberOfSets() {
		return fuzzySets.getNumberOfSets();
	}

	public FuzzySet getFuzzySet(int index) {
		return fuzzySets.getFuzzySet(index);
	}

	private void initValueDifferencesArray() {

		double sortedValues[] = actualValues.getSortedActualValues();
		double tempArray[] = new double[sortedValues.length - 1];
		for (int i = 0; i < sortedValues.length - 1; i++) {
			double value1 = sortedValues[i];
			double value2 = sortedValues[i + 1];
			double diff = Math.abs(value1 - value2);
			tempArray[i] = diff;
		}
		valueDifferences = tempArray;
	}// initValueDifferencesArray

	private void initFuzzySets() {

		Iterator<ActualValue> itr = actualValues.getActualValuesIterator();

		while (itr.hasNext()) {
			double value = itr.next().getActualValue();
			FuzzyNumber fn = fuzzyNumbers.getFuzzyNumberByValue(value);
			FuzzySet set = new TrapezoidalFuzzySet(id + "",
					(TrapezoidalFuzzyNumber) fn);
			fuzzySets.addFuzzySet(set);
			id++;
		}

	}

	/**
	 * Returns a List containing all the Fuzzy numbers
	 */
	public void initFuzzyNumbers() {

		double minBound = actualValues.getMinValue();
		double maxBound = actualValues.getMaxValue();
		double segmentLength = roundDoubleValue(0, getFunctionSpread());
		double range = (maxBound + segmentLength) - (minBound - segmentLength);
		int n = (int) (((range-segmentLength) /(2* segmentLength)) + 0.5);
		// double nn =  (range-segmentLength) /(2* segmentLength);
		// int n = (int) (((range - 1)/ (2*segmentLength))+ 0.5d);
//		double nn = (range-segmentLength)/ (2*segmentLength);
		// System.out.println("dasdfasdf "+nn);
		double roundDiff = (maxBound - minBound - segmentLength * (2 * n - 1))
				/ (2 * n - 1);
		double roundedSegmentLength = segmentLength + roundDiff;
		double lowerBound = roundDoubleValue(2, minBound - segmentLength);
		double leftCore = roundDoubleValue(2, minBound);
		double rightCore = roundDoubleValue(2, leftCore + roundedSegmentLength);
		double upperBound = roundDoubleValue(2, rightCore
				+ roundedSegmentLength);

		for (int i = 0; i < n; i++) {

			FuzzyNumber fNumber = new TrapezoidalFuzzyNumber(roundDoubleValue(
					0, lowerBound), roundDoubleValue(0, leftCore),
					roundDoubleValue(0, rightCore), roundDoubleValue(0,
							upperBound));
			fuzzyNumbers.addFuzzyNumber(fNumber); // Number is added here
			lowerBound = rightCore;
			leftCore = upperBound;
			rightCore = leftCore + roundedSegmentLength;

			if (i == n - 2)
				upperBound = rightCore + segmentLength;
			else
				upperBound = rightCore + roundedSegmentLength;
		}
	}

	private double roundDoubleValue(int numberOfDecimals, double value) {

		BigDecimal bd = new BigDecimal(value);
		bd = bd.setScale(numberOfDecimals, RoundingMode.HALF_UP);
		return bd.doubleValue();

	}

	// standard deviation of differences
	private double getStdDev() {

		double averageDiff = getAvgDiff();
		double sumSquaredDiff = 0;
		for (int i = 0; i < valueDifferences.length; i++) {
			double value = valueDifferences[i];
			sumSquaredDiff += Math.pow(value - averageDiff, 2);
		}
		double variance = sumSquaredDiff / valueDifferences.length;
		return Math.sqrt(variance);
	}

	// calculates the spread of the trapezoidal function
	private double getFunctionSpread() {

		double avgDiff = getAvgDiff();
		double stdDev = getStdDev();
		double maxLimit = avgDiff + stdDev;
		double minLimit = avgDiff - stdDev;
		double sumDiff = 0;
		double count = 0;

		for (int i = 0; i < valueDifferences.length; i++) {
			double diff = valueDifferences[i];
			if (maxLimit >= diff && minLimit <= diff) {
				sumDiff += diff;
				count++;
			}
		}
		if (count == 0) {
			System.out.println("----- average diff ---------- " + avgDiff);
			return avgDiff;
		}

		else
			return sumDiff / count;
	}

	// calculates average difference when values are sorted
	private double getAvgDiff() {

		double sum = 0;
		for (int i = 0; i < valueDifferences.length; i++) {
			double value = valueDifferences[i];
			sum += value;
		}
		return sum / valueDifferences.length;
	}

	public static void main(String[] args) {

		AlabamaEnrollmentsContainer container = new AlabamaEnrollmentsContainer();
		ActualValueContainer actualVal = container
				.getActualEnrollmentsContainer();
		TrapezoidalFuzzifier fuzzifier = new TrapezoidalFuzzifier(actualVal);

		Iterator<FuzzyNumber> itr = fuzzifier.getFuzzyNumberContainer();

		while (itr.hasNext()) {

			System.out.println(itr.next());

		}

		Iterator<FuzzySet> itr2 = fuzzifier.getFuzzySetsIterator();
		System.out.println("----------- Fuzzy Sets ------------");
		while (itr2.hasNext()) {
			System.out.println(itr2.next());

		}

	}

}
